/*
    This is a romfs file system driver for Windows NT and Windows 2000.
    Copyright (C) 1999, 2000, 2001 Bo Brantn.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ntifs.h"
#include "fsd.h"
#include "rom_fs.h"
#include "border.h"

NTSTATUS
FsdQueryVolumeInformation (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
    PDEVICE_OBJECT          DeviceObject;
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;
    PFSD_VCB                Vcb;
    PIRP                    Irp;
    PIO_STACK_LOCATION      IoStackLocation;
    FS_INFORMATION_CLASS    FsInformationClass;
    ULONG                   Length;
    PVOID                   Buffer;
    BOOLEAN                 VcbResourceAcquired = FALSE;

    __try
    {
        ASSERT(IrpContext != NULL);

        ASSERT((IrpContext->Identifier.Type == ICX) &&
               (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

        DeviceObject = IrpContext->DeviceObject;

        //
        // This request is not allowed on the main device object
        //
        if (DeviceObject == FsdGlobalData.DeviceObject)
        {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }

        Vcb = (PFSD_VCB) DeviceObject->DeviceExtension;

        ASSERT(Vcb != NULL);

        ASSERT((Vcb->Identifier.Type == VCB) &&
               (Vcb->Identifier.Size == sizeof(FSD_VCB)));

        if (!ExAcquireResourceSharedLite(
                 &Vcb->MainResource,
                 IrpContext->IsSynchronous
                 ))
        {
            Status = STATUS_PENDING;
            __leave;
        }

        VcbResourceAcquired = TRUE;

        Irp = IrpContext->Irp;

        IoStackLocation = IoGetCurrentIrpStackLocation(Irp);

        FsInformationClass =
            IoStackLocation->Parameters.QueryVolume.FsInformationClass;

        Length = IoStackLocation->Parameters.QueryVolume.Length;

        Buffer = Irp->AssociatedIrp.SystemBuffer;

        RtlZeroMemory(Buffer, Length);

        switch (FsInformationClass)
        {
        case FileFsVolumeInformation:
            {
                PFILE_FS_VOLUME_INFORMATION FileFsVolumeInformation;
                ULONG                       VolumeLabelLength;
                ULONG                       RequiredLength;

                if (Length < sizeof(FILE_FS_VOLUME_INFORMATION))
                {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }

                FileFsVolumeInformation = (PFILE_FS_VOLUME_INFORMATION) Buffer;

                FileFsVolumeInformation->VolumeCreationTime.QuadPart = 0;

                FileFsVolumeInformation->VolumeSerialNumber =
                    be32_to_cpu(Vcb->romfs_super_block->checksum);

                VolumeLabelLength = strnlen(
                                        Vcb->romfs_super_block->name,
                                        ROMFS_MAXFN
                                        );

                FileFsVolumeInformation->VolumeLabelLength =
                    VolumeLabelLength * 2;

                // I don't know what this means
                FileFsVolumeInformation->SupportsObjects = FALSE;

                RequiredLength = sizeof(FILE_FS_VOLUME_INFORMATION)
                    + VolumeLabelLength * 2 - sizeof(WCHAR);

                if (Length < RequiredLength)
                {
                    Irp->IoStatus.Information =
                        sizeof(FILE_FS_VOLUME_INFORMATION);
                    Status = STATUS_BUFFER_OVERFLOW;
                    __leave;
                }

                FsdCharToWchar(
                    FileFsVolumeInformation->VolumeLabel,
                    Vcb->romfs_super_block->name,
                    VolumeLabelLength
                    );

                Irp->IoStatus.Information = RequiredLength;
                Status = STATUS_SUCCESS;
                __leave;
            }

        case FileFsSizeInformation:
            {
                PFILE_FS_SIZE_INFORMATION FileFsSizeInformation;

                if (Length < sizeof(FILE_FS_SIZE_INFORMATION))
                {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }

                FileFsSizeInformation = (PFILE_FS_SIZE_INFORMATION) Buffer;

#ifndef FSD_RO
                if (!Vcb->ReadOnly)
                {
                    FileFsSizeInformation->TotalAllocationUnits.QuadPart =
                        Vcb->PartitionInformation.PartitionLength.QuadPart /
                        ROMBSIZE;

                    FileFsSizeInformation->AvailableAllocationUnits.QuadPart =
                        (Vcb->PartitionInformation.PartitionLength.QuadPart -
                        be32_to_cpu(Vcb->romfs_super_block->size)) / ROMBSIZE;
                }
                else
#endif // !FSD_RO
                {
                    // On a readonly filesystem total size is the size of the
                    // contents and available size is zero

                    FileFsSizeInformation->TotalAllocationUnits.QuadPart =
                        be32_to_cpu(Vcb->romfs_super_block->size) / ROMBSIZE;

                    FileFsSizeInformation->AvailableAllocationUnits.QuadPart =
                        0;
                }

                FileFsSizeInformation->SectorsPerAllocationUnit =
                    ROMBSIZE / Vcb->DiskGeometry.BytesPerSector;

                FileFsSizeInformation->BytesPerSector =
                    Vcb->DiskGeometry.BytesPerSector;

                Irp->IoStatus.Information = sizeof(FILE_FS_SIZE_INFORMATION);
                Status = STATUS_SUCCESS;
                __leave;
            }

        case FileFsDeviceInformation:
            {
                PFILE_FS_DEVICE_INFORMATION FileFsDeviceInformation;

                if (Length < sizeof(FILE_FS_DEVICE_INFORMATION))
                {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }

                FileFsDeviceInformation = (PFILE_FS_DEVICE_INFORMATION) Buffer;

                FileFsDeviceInformation->DeviceType =
                    Vcb->TargetDeviceObject->DeviceType;

                FileFsDeviceInformation->Characteristics =
                    Vcb->TargetDeviceObject->Characteristics;

#ifndef FSD_RO
                if (Vcb->ReadOnly)
#endif
                {
                    FileFsDeviceInformation->Characteristics |=
                        FILE_READ_ONLY_DEVICE;
                }

                Irp->IoStatus.Information = sizeof(FILE_FS_DEVICE_INFORMATION);
                Status = STATUS_SUCCESS;
                __leave;
            }

        case FileFsAttributeInformation:
            {
                PFILE_FS_ATTRIBUTE_INFORMATION  FileFsAttributeInformation;
                ULONG                           RequiredLength;

                if (Length < sizeof(FILE_FS_ATTRIBUTE_INFORMATION))
                {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }

                FileFsAttributeInformation =
                    (PFILE_FS_ATTRIBUTE_INFORMATION) Buffer;

                FileFsAttributeInformation->FileSystemAttributes =
                    FILE_CASE_SENSITIVE_SEARCH | FILE_CASE_PRESERVED_NAMES;

                FileFsAttributeInformation->MaximumComponentNameLength =
                    ROMFS_MAXFN;

                FileFsAttributeInformation->FileSystemNameLength =
                    sizeof(DRIVER_NAME) * 2;

                RequiredLength = sizeof(FILE_FS_ATTRIBUTE_INFORMATION) +
                    sizeof(DRIVER_NAME) * 2 - sizeof(WCHAR);

                if (Length < RequiredLength)
                {
                    Irp->IoStatus.Information =
                        sizeof(FILE_FS_ATTRIBUTE_INFORMATION);
                    Status = STATUS_BUFFER_OVERFLOW;
                    __leave;
                }

                FsdCharToWchar(
                    FileFsAttributeInformation->FileSystemName,
                    DRIVER_NAME,
                    sizeof(DRIVER_NAME)
                    );

                Irp->IoStatus.Information = RequiredLength;
                Status = STATUS_SUCCESS;
                __leave;
            }

        default:
            Status = STATUS_INVALID_INFO_CLASS;
        }
    }
    __finally
    {
        if (VcbResourceAcquired)
        {
            ExReleaseResourceForThreadLite(
                &Vcb->MainResource,
                ExGetCurrentResourceThread()
                );
        }

        if (!IrpContext->ExceptionInProgress)
        {
            if (Status == STATUS_PENDING)
            {
                FsdQueueRequest(IrpContext);
            }
            else
            {
                IrpContext->Irp->IoStatus.Status = Status;

                FsdCompleteRequest(
                    IrpContext->Irp,
                    (CCHAR)
                    (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
                    );

                FsdFreeIrpContext(IrpContext);
            }
        }
    }

    return Status;
}
